/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "utf8_io.h"

/* This variable is only used to stop compile time FORTIFY_SOURCE warnings
 * when casting a return-value-function to (void).
 * E.g (void) write(...) now becomes write_n = (long) write(...) and write_n
 * is ignored. Could use ssize_t but safer with type long.  */
long write_n;

/* function templates */
void utf8_getchar_init();
unsigned char utf8_getchar(int fd);
void utf8_print(int fd, unsigned char *buf, long buf_bytes);
void utf8_read_bom(int fd);
void utf8_print_bom(int fd);


void utf8_getchar_init()
{
  /* initialise variables for mimicking char reads from a block read */
  utf8_flag_EOF = FALSE;
  utf8_read     = 0;
  utf8_cntr     = 0;
}


unsigned char utf8_getchar(fd)
int fd;
{
  /* don't read beyond EOF more than once */
  if (utf8_flag_EOF)
    return(0);

  /* Read a block of input and return one char at a time */
  if (utf8_cntr >= utf8_read)
  {
    utf8_read = read(fd, utf8_buf, BUFSIZE);
    utf8_cntr = 0;
    if (utf8_read < 1)
      utf8_flag_EOF = TRUE;
  }

  return(utf8_buf[utf8_cntr++]);
}


void utf8_print(fd, buf, buf_bytes)
int fd;
unsigned char *buf;
long buf_bytes;
{
  /* print UTF-8 characters */
  /* printf() won't work here because of the need to print the null character */
  long i;

  /* output in manageable block sizes */
  for (i = 0 ; i < buf_bytes / BUFSIZE ; i++)
    write_n = (long) write(fd, &buf[i * BUFSIZE], BUFSIZE);
  i = buf_bytes % BUFSIZE;
  if (i > 0)
    write_n = (long) write(fd, &buf[buf_bytes - i], i);
}


void utf8_read_bom(fd)
int fd;
{
  /* Pre-load the read block buffer and by-pass the BOM if any. */
  /* This must be the first read and is only called once. */

  /* Read a block of input and return one char at a time */
  if (!utf8_flag_EOF && utf8_cntr >= utf8_read)
  {
    utf8_read = read(fd, utf8_buf, BUFSIZE);
    if (utf8_read < 1)
      utf8_flag_EOF = TRUE;

    if (utf8_read > 2 && xstrncmp((unsigned char *) UTF8_BOM, utf8_buf, 3) == 0)
      utf8_cntr = 3;
    else
      utf8_cntr = 0;
  }
}


void utf8_print_bom(fd)
int fd;
{
  /* print UTF-8 BOM */
  write_n = (long) write(fd, UTF8_BOM, xstrlen((unsigned char *) UTF8_BOM));
}
