/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "validate_args.h"

/* function templates */
int  process_arg_column(unsigned char *s);


int process_arg_column(s)
unsigned char *s;
{
  /* translate column arg to column structure */
  unsigned char *column_range, *column_name;
  int s_length, r_length, n_length;
  int i, j, ret;
  char value[10];
  int i_value;

  /* initialise */
  ret      = TRUE;
  s_length = xstrlen(s);

  if (s_length == 0)
    ret = FALSE;

  /* split arg into range/name components */
  i = 0;
  while (i < s_length && s[i] != ':')
    i++;
  if (i == s_length)
  {
    column_range = s;
    r_length     = s_length;
    column_name  = NULL;
    n_length     = 0;
  }
  else
  {
    column_range = s;
    r_length     = i;
    column_name  = &s[i + 1];
    n_length     = s_length - i - 1;
  }

  /* ensure legal range chars */
  for (i = 0 ; i < r_length ; i++)
  {
    switch (column_range[i])
    {
      case '0' : 
      case '1' : 
      case '2' : 
      case '3' : 
      case '4' : 
      case '5' : 
      case '6' : 
      case '7' : 
      case '8' : 
      case '9' : 
      case '$' : 
      case '-' : 
      case '\'' : 
      case '\"' : 
                 break;

      default :  ret = FALSE;;
    }
  }

  if (ret)
  {
    /* allocate memory for expel structure */
    if (ncolumn)
      column = (struct_column *) realloc(column, ((ncolumn + 1) * sizeof(struct_column)));
    else
      column = (struct_column *) malloc(sizeof(struct_column));
    if (column == NULL) out_of_mem();

    column[ncolumn].name = column_name;
    column[ncolumn].min  = RANGE_UNSET;
    column[ncolumn].max  = RANGE_UNSET;

    /* populate expel structure */
    j = 0;
    i_value = 0;
    value[0] = '\0';
    for (i = 0 ; i < r_length ; i++)
    {
      switch (column_range[i])
      {
        case '\'' :
        case '\"' : i_value = 0;
                    break;

        case '$' : if (column[ncolumn].min == RANGE_UNSET)
                   {
                     column[ncolumn].min = -1;
                     column[ncolumn].max = -1;
                   }
                   else
                     column[ncolumn].max = -1;
                   i_value = 0;
                   break;

        case '-' : if (i_value)
                   {
                     if (column[ncolumn].min == RANGE_UNSET)
                     {
                       column[ncolumn].min = atol(value) - 1;
                       if (column[ncolumn].min < 0)
                         ret = FALSE;
                     }
                   }
                   else if (column[ncolumn].min == RANGE_UNSET)
                     column[ncolumn].min = 0;
                   i_value = 0;
                   /* pre-populate max field in case not supplied */
                   column[ncolumn].max = -1;
                   break;

        default :  value[i_value++] = column_range[i];
                   value[i_value]   = '\0';
      }
    }
    if (i_value)
    {
      if (column[ncolumn].min == RANGE_UNSET)
      {
        column[ncolumn].min = atol(value) - 1;
        if (column[ncolumn].min < 0)
          ret = FALSE;
        /* pre-populate max field in case not supplied */
        column[ncolumn].max = column[ncolumn].min;
      }
      else
      {
        column[ncolumn].max = atol(value) - 1;
        if (column[ncolumn].max < 0)
          ret = FALSE;
      }
    }
    /* quick sanity check */
    if (column[ncolumn].max > 0 && column[ncolumn].min > column[ncolumn].max)
      ret = FALSE;
    if (column[ncolumn].min < 0 && column[ncolumn].max >= 0)
      ret = FALSE;

    ncolumn++;
  }

  return(ret);
}

