/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

long i_curr_field;

/* function templates */
void process_output_record_default(long record_number);
void process_output_record_kvp(long record_number);
void process_output_record_schema(long record_number);
void process_schema_level(int level);
void mkelem(int level, unsigned char *t, int t_length, unsigned char *f, long f_length, long n, int sw_numbered_tag, int sw_numbered_attr, int level_d, int d, int sw_delimiter_attr, int sw_level_tag, int sw_level_attr, int sw_timestamp_attr, int sw_custom_attr);
void mktag(int tag_control, int level, unsigned char *t, int t_length, long n, int sw_numbered_tag, int sw_numbered_attr, int level_d, int d, int sw_delimiter_attr, int sw_level_tag, int sw_level_attr, int sw_timestamp_attr, int sw_custom_attr);
void make_timestamp_array(time_t t, int type);
void append_delimiter_string(int level, int d);
void strcpy_output_record(int level, const unsigned char *s_src);
void strncpy_output_record(const unsigned char *s_src, long n);
void strcat_output_record(const unsigned char *s_src);
void strncat_output_record(const unsigned char *s_src, long n);
void format_key_value_tagname(long i);


void process_output(record_number)
long record_number;
{
  /* print an output record in XML format */

  if (sw_blank_lines || input_record.nfields)
  {
    /* print an output record in the requested format */
    if (sw_schema == SCHEMA_DEFAULT)
    {
      if (!sw_key_value_pairs)
        process_output_record_default(record_number);
      else
        process_output_record_kvp(record_number);
    }
    else
      process_output_record_schema(record_number);

    /* increment record printed counter */
    record_cntr_printed++;
  }
  else
    record_cntr_blank++;
}


void process_output_record_default(record_number)
long record_number;
{
  /* print an output record in default XML format */
  long i;
  int current_level;

  /* initialise counters */
  for (i = 0 ; i < nlevels ; i++)
    field_cntr[i] = 0;
  current_level = 2;

  /* open record tag */
  mktag(TAG_OPEN,
    1,
    sw_print[1]->tagname, 
    sw_print[1]->tagname_length, 
    record_number,
    sw_print[1]->numbered_tag,
    sw_print[1]->numbered_attr,
    1,
    delimiter[1]->match_d,
    sw_print[1]->delimiter_attr,
    sw_print[1]->level_tag,
    sw_print[1]->level_attr,
    sw_print[1]->timestamp_attr,
    nattrname);
  utf8_print(STDOUT_FD, output_record, output_record_length);

  /* print a line tag */
  if (sw_line_tags) 
  {
    (void) sprintf(temp_array, "<%s>%ld</%s>\n", schema_linenumber, record_number, schema_linenumber);
    strcpy_output_record(2, (unsigned char *) temp_array);
    utf8_print(STDOUT_FD, output_record, output_record_length);
  }

  for (i = 0 ; i < input_record.nfields ; i++)
  {
    while (input_record.field[i]->level != current_level)
    {
      if (input_record.field[i]->level > current_level)
      {
        /* print tree of opening field tags */
        mktag(TAG_OPEN,
          current_level,
          sw_print[current_level]->tagname, 
          sw_print[current_level]->tagname_length, 
          field_cntr[current_level] + 1,
          sw_print[current_level]->numbered_tag,
          sw_print[current_level]->numbered_attr,
          input_record.field[i - 1]->level_d,
          input_record.field[i - 1]->match_d,
          sw_print[current_level]->delimiter_attr,
          sw_print[current_level]->level_tag,
          sw_print[current_level]->level_attr,
          sw_print[current_level]->timestamp_attr,
          nattrname);
        utf8_print(STDOUT_FD, output_record, output_record_length);
        current_level++;
        /* initialise next deeper level numbering */
        if (current_level < nlevels)
          field_cntr[current_level] = 0;
      }
      else if (input_record.field[i]->level < current_level)
      {
        /* print tree of closing field tags */
        current_level--;
        mktag(TAG_CLOSE,
          current_level,
          sw_print[current_level]->tagname, 
          sw_print[current_level]->tagname_length, 
          field_cntr[current_level] + 1,
          sw_print[current_level]->numbered_tag,
          0,
          0,
          DELIMITER_INDEX_UNMATCHED,
          0,
          sw_print[current_level]->level_tag,
          0,
          0,
          0);
        utf8_print(STDOUT_FD, output_record, output_record_length);
        field_cntr[current_level] += 1;
      }
    }

    /* print a complete field element at its depth level */
    if (input_record.field[i]->type == FIELD_CDATA)
    {
      mkelem(input_record.field[i]->level,
        input_record.field[i]->tagname, 
        xstrlen(input_record.field[i]->tagname), 
        &input_record.line[input_record.field[i]->start], 
        input_record.field[i]->length,
        field_cntr[input_record.field[i]->level] + 1,
        sw_print[input_record.field[i]->level]->numbered_tag,
        sw_print[input_record.field[i]->level]->numbered_attr,
        input_record.field[i]->level_d,
        input_record.field[i]->match_d,
        sw_print[input_record.field[i]->level]->delimiter_attr,
        sw_print[input_record.field[i]->level]->level_tag,
        sw_print[input_record.field[i]->level]->level_attr,
        sw_print[input_record.field[i]->level]->timestamp_attr,
        nattrname);
      utf8_print(STDOUT_FD, output_record, output_record_length);
      field_cntr[input_record.field[i]->level] += 1;
    }
  }

  /* print tree of closing field tags */
  while (current_level > 2)
  {
    current_level--;
    mktag(TAG_CLOSE,
      current_level,
      sw_print[current_level]->tagname, 
      sw_print[current_level]->tagname_length, 
      field_cntr[current_level] + 1,
      sw_print[current_level]->numbered_tag,
      0,
      0,
      DELIMITER_INDEX_UNMATCHED,
      0,
      sw_print[current_level]->level_tag,
      0,
      0,
      0);
    utf8_print(STDOUT_FD, output_record, output_record_length);
    field_cntr[current_level] += 1;
  }
 
  /* close record tag */
  mktag(TAG_CLOSE,
    1,
    sw_print[1]->tagname, 
    sw_print[1]->tagname_length, 
    record_number,
    sw_print[1]->numbered_tag,
    0,
    0,
    DELIMITER_INDEX_UNMATCHED,
    0,
    sw_print[1]->level_tag,
    0,
    0,
    0);
  utf8_print(STDOUT_FD, output_record, output_record_length);
}


void process_output_record_kvp(record_number)
long record_number;
{
  /* print an output record in key/value pairs XML format */
  long i, j, k;
  int current_level, record_type;
  int flag_drop_level;

  /* initialise */
  for (current_level = 0 ; current_level < nlevels ; current_level++)
  {
    field_cntr[current_level]   = 0;
    kvp_tag[current_level]->set = FALSE;
  }
  field_cntr[1] = record_number - 1;

  /* itentify FIELD_KEYTAG_* candidates */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    if (input_record.field[i]->type == FIELD_CDATA)
    {
      /* check if restricted to specified levels */
      if (!sw_kvp_levels_all)
      {
        if (!kvp_tag[input_record.field[i]->level]->enabled)
          continue;
      }

      /* blank fields don't make good tags so coerce them out */
      if (!sw_blank_lines && !input_record.field[i]->length)
      {
        input_record.field[i]->type = FIELD_SUBRECORD;
        continue;
      }

      /* look ahead */
      k = 0;
      for (j = i + 1 ; j < input_record.nfields ; j++)
      {
        if (k < 2 && input_record.field[j]->level >= input_record.field[i]->level)
        {
          if (input_record.field[j]->type == FIELD_CDATA && kvp_tag[input_record.field[j]->level]->set == FALSE)
            k++;
          /* unset the kvp flag */
          if (input_record.field[j]->level > input_record.field[i]->level)
            kvp_tag[input_record.field[i]->level]->set = FALSE;
        }
        else
          break;
      } /* end for/j */

      if (k == 1)
      {
        input_record.field[i]->type = FIELD_KEYTAG_ELEMENT;
        format_key_value_tagname(i);

        i = j;
      }
      else if (k > 1)
      {
        if (!kvp_tag[input_record.field[i]->level]->set)
        {
          input_record.field[i]->type = FIELD_KEYTAG_RECORD;
          format_key_value_tagname(i);

          kvp_tag[input_record.field[i]->level]->set = TRUE;
          for (current_level = input_record.field[i]->level + 1 ; current_level < nlevels ; current_level++)
            kvp_tag[current_level]->set = FALSE;
        }

        j = i + 1;
        while (j < input_record.nfields && input_record.field[j]->level == input_record.field[i]->level 
            && (input_record.field[j]->type == FIELD_CDATA || input_record.field[j]->type == FIELD_EXPELLED))
          j++;
        i = j - 1;
      }
    }
    else if (input_record.field[i]->type == FIELD_SUBRECORD)
    {
      for (current_level = input_record.field[i]->level + 1 ; current_level < nlevels ; current_level++)
        kvp_tag[current_level]->set = FALSE;
    }
  } /* end for/i */

  /* adjust levels to reflect the revised field layout */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    if (input_record.field[i]->type == FIELD_KEYTAG_ELEMENT)
    {
      /* backward update */
      j = i - 1;
      while (j >= 0 && input_record.field[j]->level > 1
          && ((input_record.field[j]->level <= input_record.field[i]->level && input_record.field[j]->type == FIELD_SUBRECORD)
          || (input_record.field[j]->level == input_record.field[i]->level && input_record.field[j]->type == FIELD_EXPELLED)))
        input_record.field[j--]->level--;
      /* forward update */
      input_record.field[i]->level--;
      i++;
      if (i < input_record.nfields)
        input_record.field[i]->level--;
    }
    else if (input_record.field[i]->type == FIELD_KEYTAG_RECORD)
    {
      /* check if 2 consecutive superfluous levels exist */
      flag_drop_level = TRUE;
      for (j = i + 1; flag_drop_level && j < input_record.nfields && input_record.field[j]->level >= input_record.field[i]->level ; j++)
      {
        if (input_record.field[j]->level == input_record.field[i]->level || input_record.field[j]->level == input_record.field[i]->level + 1) 
          if (input_record.field[j]->type == FIELD_CDATA || input_record.field[j]->type == FIELD_KEYTAG_ELEMENT || input_record.field[j]->type == FIELD_KEYTAG_RECORD)
            flag_drop_level = FALSE;
      }
          
      if (flag_drop_level)
      {
          /* remove the lower superfluous level that has no fields as it messes up the key/pair output format */
          for (k = i + 1 ; k < j ; k++)
            input_record.field[k]->level--;

          input_record.field[i]->level--;
      }
      else
      {
        /* backward update */
        j = i - 1;
        while (j >= 0 && input_record.field[j]->level > 1 
            && ((input_record.field[j]->level <= input_record.field[i]->level && input_record.field[j]->type == FIELD_SUBRECORD)
            || (input_record.field[j]->level == input_record.field[i]->level && input_record.field[j]->type == FIELD_EXPELLED)))
          input_record.field[j--]->level--;
        /* forward update */
        j = i + 1;
        while (j < input_record.nfields && input_record.field[j]->level == input_record.field[i]->level
            && (input_record.field[j]->type == FIELD_CDATA || input_record.field[j]->type == FIELD_EXPELLED))
           j++;
        input_record.field[i]->level--;
        i = j - 1;
      }
    }
  } /* end for/i */

  /* 
   * output adjusted fields of this record 
   */

  kvp_tag[1]->name   = sw_print[1]->tagname; 
  kvp_tag[1]->length = sw_print[1]->tagname_length; 
  kvp_tag[1]->level_d = 1;
  kvp_tag[1]->match_d = delimiter[1]->match_d;
  record_type = FIELD_KEYTAG_RECORD;

  /* find the first field for becoming the key of key/value */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    if (input_record.field[i]->type == FIELD_SUBRECORD || input_record.field[i]->type == FIELD_EXPELLED)
      continue;

    if (input_record.field[i]->level == 1)
    {
      /* set the parent tag name of this level to be the text of the first field of the current level */
      if (input_record.field[i]->type == FIELD_KEYTAG_ELEMENT || input_record.field[i]->type == FIELD_KEYTAG_RECORD)
      {
        /* set the key */
        if (input_record.field[i]->length)
        {
          kvp_tag[1]->name   = &input_record.line[input_record.field[i]->start];
          kvp_tag[1]->length = input_record.field[i]->length;
        }
        else
        {
          kvp_tag[1]->name   = sw_print[1]->tagname;
          kvp_tag[1]->length = sw_print[1]->tagname_length;
        }
        kvp_tag[1]->level_d = input_record.field[i]->level_d;
        kvp_tag[1]->match_d = input_record.field[i]->match_d;
        record_type         = input_record.field[i]->type;
      }
    }
    /* if the first valid field is not a tag then use the default */
    break;
  } /* end for/i */

  if (record_type != FIELD_KEYTAG_ELEMENT)
  {
    /* open record tag */
    mktag(TAG_OPEN,
      1,
      kvp_tag[1]->name, 
      kvp_tag[1]->length, 
      record_number,
      sw_print[1]->numbered_tag,
      sw_print[1]->numbered_attr,
      kvp_tag[1]->level_d,
      kvp_tag[1]->match_d,
      sw_print[1]->delimiter_attr,
      sw_print[1]->level_tag,
      sw_print[1]->level_attr,
      sw_print[1]->timestamp_attr,
      nattrname);
    utf8_print(STDOUT_FD, output_record, output_record_length);

    /* print a line tag */
    if (sw_line_tags) 
    {
      (void) sprintf(temp_array, "<%s>%ld</%s>\n", schema_linenumber, record_number, schema_linenumber);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(STDOUT_FD, output_record, output_record_length);
    }
  
    current_level = 2;
    for (i = 0 ; i < input_record.nfields ; i++)
    {
      if (input_record.field[i]->level < 2 || input_record.field[i]->type == FIELD_EXPELLED)
        continue;

      if (input_record.field[i]->level < current_level)
      {
        while (input_record.field[i]->level != current_level)
        {
          /* print tree of closing field tags */
          current_level--;
          mktag(TAG_CLOSE,
            current_level,
            kvp_tag[current_level]->name, 
            kvp_tag[current_level]->length, 
            field_cntr[current_level] + 1,
            sw_print[current_level]->numbered_tag,
            0,
            0,
            DELIMITER_INDEX_UNMATCHED,
            0,
            sw_print[current_level]->level_tag,
            0,
            0,
            0);
          utf8_print(STDOUT_FD, output_record, output_record_length);
          field_cntr[current_level] += 1;
          /* reset tag name */
          kvp_tag[current_level + 1]->name   = sw_print[current_level + 1]->tagname; 
          kvp_tag[current_level + 1]->length = sw_print[current_level + 1]->tagname_length; 
          if (current_level == 2 && (input_record.field[i]->type == FIELD_CDATA || input_record.field[i]->type == FIELD_SUBRECORD))
          {
            kvp_tag[current_level]->name   = sw_print[current_level]->tagname; 
            kvp_tag[current_level]->length = sw_print[current_level]->tagname_length; 
          }
        }
      }

      if (input_record.field[i]->type == FIELD_KEYTAG_RECORD || input_record.field[i]->type == FIELD_KEYTAG_ELEMENT)
      {
        /* set the key */
        if (input_record.field[i]->length)
        {
          kvp_tag[input_record.field[i]->level]->name   = &input_record.line[input_record.field[i]->start];
          kvp_tag[input_record.field[i]->level]->length = input_record.field[i]->length;
        }
        else
        {
          kvp_tag[input_record.field[i]->level]->name   = sw_print[input_record.field[i]->level]->tagname;
          kvp_tag[input_record.field[i]->level]->length = sw_print[input_record.field[i]->level]->tagname_length;
        }
        kvp_tag[input_record.field[i]->level]->level_d    = input_record.field[i]->level_d;
        kvp_tag[input_record.field[i]->level]->match_d    = input_record.field[i]->match_d;
        if (input_record.field[i]->type == FIELD_KEYTAG_RECORD && input_record.field[i]->level + 1 < nlevels)
        {
          kvp_tag[input_record.field[i]->level + 1]->name   = sw_print[input_record.field[i]->level + 1]->tagname; 
          kvp_tag[input_record.field[i]->level + 1]->length = sw_print[input_record.field[i]->level + 1]->tagname_length; 
        }
      }

      if (input_record.field[i]->level > current_level)
      {
        while (input_record.field[i]->level != current_level)
        {
          /* print tree of opening field tags */
          if (input_record.field[i]->type == FIELD_CDATA || input_record.field[i]->type == FIELD_SUBRECORD)
          {
            /* set correct delimiter for this tag */
            kvp_tag[current_level]->level_d = input_record.field[i - 1]->level_d;
            kvp_tag[current_level]->match_d = input_record.field[i - 1]->match_d;
          }
          mktag(TAG_OPEN,
            current_level,
            kvp_tag[current_level]->name, 
            kvp_tag[current_level]->length, 
            field_cntr[current_level] + 1,
            sw_print[current_level]->numbered_tag,
            sw_print[current_level]->numbered_attr,
            kvp_tag[current_level]->level_d,
            kvp_tag[current_level]->match_d,
            sw_print[current_level]->delimiter_attr,
            sw_print[current_level]->level_tag,
            sw_print[current_level]->level_attr,
            sw_print[current_level]->timestamp_attr,
            nattrname);
          utf8_print(STDOUT_FD, output_record, output_record_length);
          current_level++;
          /* initialise next deeper level numbering */
          if (current_level < nlevels)
            field_cntr[current_level] = 0;
        }
      }

      /* print a complete field element at its depth level */
      if (input_record.field[i]->type == FIELD_CDATA)
      {
        if (sw_blank_lines || input_record.field[i]->length > 0 || kvp_tag[input_record.field[i]->level]->name != sw_print[input_record.field[i]->level]->tagname)
        {
          mkelem(input_record.field[i]->level,
            kvp_tag[input_record.field[i]->level]->name, 
            kvp_tag[input_record.field[i]->level]->length, 
            &input_record.line[input_record.field[i]->start], 
            input_record.field[i]->length,
            field_cntr[input_record.field[i]->level] + 1,
            sw_print[input_record.field[i]->level]->numbered_tag,
            sw_print[input_record.field[i]->level]->numbered_attr,
            input_record.field[i]->level_d,
            input_record.field[i]->match_d,
            sw_print[input_record.field[i]->level]->delimiter_attr,
            sw_print[input_record.field[i]->level]->level_tag,
            sw_print[input_record.field[i]->level]->level_attr,
            sw_print[input_record.field[i]->level]->timestamp_attr,
            nattrname);
          utf8_print(STDOUT_FD, output_record, output_record_length);
          field_cntr[input_record.field[i]->level] += 1;

          if (current_level == 2)
          {
            /* reset tag name */
            kvp_tag[current_level]->name   = sw_print[current_level]->tagname; 
            kvp_tag[current_level]->length = sw_print[current_level]->tagname_length; 
          }
        }
      }
    } /* end for/i */

    /* print tree of closing field tags */
    while (current_level > 2)
    {
      current_level--;
      mktag(TAG_CLOSE,
        current_level,
        kvp_tag[current_level]->name, 
        kvp_tag[current_level]->length, 
        field_cntr[current_level] + 1,
        sw_print[current_level]->numbered_tag,
        0,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        sw_print[current_level]->level_tag,
        0,
        0,
        0);
      utf8_print(STDOUT_FD, output_record, output_record_length);
      field_cntr[current_level] += 1;
      /* reset tag name */
      kvp_tag[current_level + 1]->name   = sw_print[current_level + 1]->tagname; 
      kvp_tag[current_level + 1]->length = sw_print[current_level + 1]->tagname_length; 
    }
    /* reset level 2 tag name */
    kvp_tag[2]->name   = sw_print[2]->tagname; 
    kvp_tag[2]->length = sw_print[2]->tagname_length; 
  
    /* close record tag */
    mktag(TAG_CLOSE,
      1,
      kvp_tag[1]->name, 
      kvp_tag[1]->length, 
      record_number,
      sw_print[1]->numbered_tag,
      0,
      0,
      DELIMITER_INDEX_UNMATCHED,
      0,
      sw_print[1]->level_tag,
      0,
      0,
      0);
    utf8_print(STDOUT_FD, output_record, output_record_length);
  }
  else  /* record_type = FIELD_KEYTAG_ELEMENT */
  {
    for (i = 0 ; i < input_record.nfields ; i++)
    {
      if (input_record.field[i]->type == FIELD_KEYTAG_ELEMENT)
      {
        if (input_record.field[i]->length)
        {
          kvp_tag[input_record.field[i]->level]->name   = &input_record.line[input_record.field[i]->start];
          kvp_tag[input_record.field[i]->level]->length = input_record.field[i]->length;
        }
        else
        {
          kvp_tag[input_record.field[i]->level]->name    = sw_print[input_record.field[i]->level]->tagname;
          kvp_tag[input_record.field[i]->level]->length  = sw_print[input_record.field[i]->level]->tagname_length;
        }
        kvp_tag[input_record.field[i]->level]->level_d = input_record.field[i]->level_d;
        kvp_tag[input_record.field[i]->level]->match_d = input_record.field[i]->match_d;
      }
      else if (input_record.field[i]->type == FIELD_CDATA)
      {
        mkelem(input_record.field[i]->level,
          kvp_tag[input_record.field[i]->level]->name, 
          kvp_tag[input_record.field[i]->level]->length, 
          &input_record.line[input_record.field[i]->start], 
          input_record.field[i]->length,
          field_cntr[input_record.field[i]->level] + 1,
          sw_print[input_record.field[i]->level]->numbered_tag,
          sw_print[input_record.field[i]->level]->numbered_attr,
          kvp_tag[input_record.field[i]->level]->level_d,
          kvp_tag[input_record.field[i]->level]->match_d,
          sw_print[input_record.field[i]->level]->delimiter_attr,
          sw_print[input_record.field[i]->level]->level_tag,
          sw_print[input_record.field[i]->level]->level_attr,
          sw_print[input_record.field[i]->level]->timestamp_attr,
          nattrname);
        utf8_print(STDOUT_FD, output_record, output_record_length);
        field_cntr[input_record.field[i]->level] += 1;
      }
    } /* end for/i */
  } /* end if/record_type */
}


void process_output_record_schema(record_number)
long record_number;
{
  /* print an output record using schema for XML format */

  /* open schema record tag */
  mktag(TAG_OPEN,
    1,
    schema_line_name, 
    xstrlen(schema_line_name), 
    record_number,
    sw_print[1]->numbered_tag,
    sw_print[1]->numbered_attr,
    1,
    delimiter[1]->match_d,
    sw_print[1]->delimiter_attr,
    sw_print[1]->level_tag,
    sw_print[1]->level_attr,
    sw_print[1]->timestamp_attr,
    nattrname);
  utf8_print(STDOUT_FD, output_record, output_record_length);

  /* print a line tag */
  if (sw_line_tags) 
  {
    (void) sprintf(temp_array, "<%s>%ld</%s>\n", schema_linenumber, record_number, schema_linenumber);
    strcpy_output_record(2, (unsigned char *) temp_array);
    utf8_print(STDOUT_FD, output_record, output_record_length);
  }

  /* print schema tree tags and data */
  i_curr_field = 0;
  process_schema_level(2);

  /* close schema record tag */
  mktag(TAG_CLOSE,
    1,
    schema_line_name, 
    xstrlen(schema_line_name), 
    record_number,
    sw_print[1]->numbered_tag,
    0,
    0,
    DELIMITER_INDEX_UNMATCHED,
    0,
    sw_print[1]->level_tag,
    0,
    0,
    0);
  utf8_print(STDOUT_FD, output_record, output_record_length);
}


void process_schema_level(level)
int level;
{
  /* process the input fields that now have an association with the schema elements */
  long i, j, f_cntr, prev_sequence;
  int flag_process_deeper;

  prev_sequence = -1;
  flag_process_deeper = TRUE;

  /* we use i to capture the tree hierachy for the tag closure in the recursion */
  i = i_curr_field;

  while (i_curr_field < input_record.nfields_cdata && flag_process_deeper)
  {
    if (input_record.field[i_curr_field]->level == level && input_record.field[i_curr_field]->sequence > prev_sequence)
    {
      prev_sequence = input_record.field[i_curr_field]->sequence;

      /* re-calculate field number as fields may occur in punctuated sequences */
      f_cntr = 0;
      if (sw_print[level]->numbered_tag || sw_print[level]->numbered_attr)
      {
        for (j = i_curr_field ; j >= 0 && input_record.field[j]->level >= level ; j--) 
        {
          if (input_record.field[j]->level == level) 
          {
            if (xstrcmp(input_record.field[i_curr_field]->tagname, input_record.field[j]->tagname) == 0)
              f_cntr++;
            if (input_record.field[j]->sequence < 1)
               break; /* break for/j */
          }
        } /* end for/j */
      }

      mkelem(input_record.field[i_curr_field]->level,
        input_record.field[i_curr_field]->tagname, 
        xstrlen(input_record.field[i_curr_field]->tagname), 
        &input_record.line[input_record.field[i_curr_field]->start], 
        input_record.field[i_curr_field]->length,
        f_cntr,
        sw_print[input_record.field[i_curr_field]->level]->numbered_tag,
        sw_print[input_record.field[i_curr_field]->level]->numbered_attr,
        input_record.field[i_curr_field]->level_d,
        input_record.field[i_curr_field]->match_d,
        sw_print[input_record.field[i_curr_field]->level]->delimiter_attr,
        sw_print[input_record.field[i_curr_field]->level]->level_tag,
        sw_print[input_record.field[i_curr_field]->level]->level_attr,
        sw_print[input_record.field[i_curr_field]->level]->timestamp_attr,
        nattrname);
      utf8_print(STDOUT_FD, output_record, output_record_length);

      i_curr_field++;
      i = i_curr_field;
    }
    else if (input_record.field[i_curr_field]->level > level)
    {
      /* re-calculate field number as fields may occur in punctuated sequences */
      f_cntr = 0;
      if (sw_print[level]->numbered_tag || sw_print[level]->numbered_attr)
      {
        for (j = i ; j >= 0 && input_record.field[j]->level > level ; j--) 
          if (input_record.field[i]->level == input_record.field[j]->level && input_record.field[j]->sequence == 0) 
            if (input_record.field[i]->schema_xpath[level] == input_record.field[j]->schema_xpath[level])
              f_cntr++;
      }

      /* print opening field tag */
      /* NOTE: Tree delimiters are meaningless with flattened input field structure
       * as they would have exactly the same value as the field delimiter.
       *   input_record.field[i]->level_d,
       *   input_record.field[i]->match_d,
       *   sw_print[level]->delimiter_attr, 
       * The above is what would have been used if they weren't, but it is the 
       * schema that dictates the tree structure not the delimiter.
       */
      mktag(TAG_OPEN,
        level,
        schema_element[input_record.field[i]->schema_xpath[level]].name, 
        xstrlen(schema_element[input_record.field[i]->schema_xpath[level]].name), 
        f_cntr,
        sw_print[level]->numbered_tag,
        sw_print[level]->numbered_attr,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        sw_print[level]->level_tag,
        sw_print[level]->level_attr,
        sw_print[level]->timestamp_attr,
        nattrname);
      utf8_print(STDOUT_FD, output_record, output_record_length);

      process_schema_level(level + 1);

      /* print closing field tag */
      mktag(TAG_CLOSE,
        level,
        schema_element[input_record.field[i]->schema_xpath[level]].name, 
        xstrlen(schema_element[input_record.field[i]->schema_xpath[level]].name), 
        f_cntr,
        sw_print[level]->numbered_tag,
        0,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        sw_print[level]->level_tag,
        0,
        0,
        0);
      utf8_print(STDOUT_FD, output_record, output_record_length);

      i = i_curr_field;
    }
    else
      flag_process_deeper = FALSE;
  }
}


void mkelem(level, t, t_length, f, f_length, n, sw_numbered_tag, sw_numbered_attr, level_d, d, sw_delimiter_attr, sw_level_tag, sw_level_attr, sw_timestamp_attr, sw_custom_attr)
int level;
unsigned char *t;
int t_length;
unsigned char *f;
long f_length;
long n;
int sw_numbered_tag;
int sw_numbered_attr;
int level_d;
int d;
int sw_delimiter_attr;
int sw_level_tag;
int sw_level_attr;
int sw_timestamp_attr;
int sw_custom_attr;
{
  /* Construct an element tag array suitable for XML and UTF-8 printing */
  /* Array output_record and output_record_length are updated */
  long i;

  output_record_length = 0;

  /* space padding */
  if (sw_xml_format == SW_SOAP)
  {
    for (i = 0 ; i < 4 ; i++)
      output_record[output_record_length++] = ' ';
  }
  for (i = 0 ; i < level ; i++)
  {
    output_record[output_record_length++] = ' ';
    output_record[output_record_length++] = ' ';
  }

  /* open element tag */
  output_record[output_record_length++] = '<';
  if (sw_xml_format == SW_SOAP)
  {
    strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
    output_record[output_record_length++] = ':';
  }
  if (sw_level_tag)
  {
    (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
    strcat_output_record((unsigned char *) temp_array);
  }
  strncat_output_record(t, t_length);
  if (n > 0)
  {
    if (sw_numbered_tag)
    {
      (void) sprintf(temp_array, "%ld", n);
      strcat_output_record((unsigned char *) temp_array);
    }
    if (sw_numbered_attr)
    {
      (void) sprintf(temp_array, " %s=\"%ld\"", DEFAULT_STR_ATTRIBUTE_NUMBER, n);
      strcat_output_record((unsigned char *) temp_array);
    }
  }
  if (sw_level_attr)
  {
    (void) sprintf(temp_array, " %s=\"%d\"", DEFAULT_STR_ATTRIBUTE_LEVEL, level);
    strcat_output_record((unsigned char *) temp_array);
  }
  if (sw_delimiter_attr)
  {
    output_record[output_record_length] = '\0';
    if (d >= 0)
    {
      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
      append_delimiter_string(level_d, d);
      output_record[output_record_length++] = '\"';
    }
    else if (d == DELIMITER_INDEX_WHITESPACE)
    {
      (void) sprintf(temp_array, " %s=\"WHITESPACE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
    }
    else if (d == DELIMITER_INDEX_NEWLINE)
    {
      (void) sprintf(temp_array, " %s=\"NEWLINE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
    }
    else if (d == DELIMITER_INDEX_COLUMN)
    {
      if (column[level_d].min > -1)
        (void) sprintf(temp_array, " %s=\"%ld", DEFAULT_STR_ATTRIBUTE_DELIMITER, column[level_d].min + 1);
      else
        (void) sprintf(temp_array, " %s=\"LAST", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
      if (column[level_d].max > -1)
        (void) sprintf(temp_array, "-%ld\"", column[level_d].max + 1);
      else
        (void) sprintf(temp_array, "-LAST\"");
      strcat_output_record((unsigned char *) temp_array);
    }
    else
    {
      (void) sprintf(temp_array, " %s=\"\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
    }
  }
  if (sw_timestamp_attr)
  {
    time(&ts);

    (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_DATE);
    strcat_output_record((unsigned char *) temp_array);
    make_timestamp_array(ts, TIMESTAMP_DATE);
    strcat_output_record((unsigned char *) temp_array);
    output_record[output_record_length++] = '\"';

    (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_SECS);
    strcat_output_record((unsigned char *) temp_array);
    make_timestamp_array(ts, TIMESTAMP_SECS);
    strcat_output_record((unsigned char *) temp_array);
    output_record[output_record_length++] = '\"';
  }
  if (sw_custom_attr)
  {
    for (i = 0 ; i < nattrname ; i++)
    {
      if (attrname[i].level < 0 || attrname[i].level == level)
      {
        output_record[output_record_length++] = ' ';
        strncat_output_record(attrname[i].name, attrname[i].name_length);
        output_record[output_record_length++] = '=';
        output_record[output_record_length++] = '\"';
        strncat_output_record(attrname[i].value, attrname[i].value_length);
        output_record[output_record_length++] = '\"';
      }
    }
  }
  output_record[output_record_length++] = '>';

  /* ============================================================
   * NOTE: from this point do not use clib routines which don't 
   * support UTF-8 on the output_record array directly.
   * ============================================================ */

  /* field */

  /* convert XML chars */
  for (i = 0 ; i < f_length ; i++)
  {
    switch (f[i]) 
    {
      /* XML characters that need escaping */
      case '<':  output_record[output_record_length++] = '&';
                 output_record[output_record_length++] = 'l';
                 output_record[output_record_length++] = 't';
                 output_record[output_record_length++] = ';';
                 break;
        
      case '>':  output_record[output_record_length++] = '&';
                 output_record[output_record_length++] = 'g';
                 output_record[output_record_length++] = 't';
                 output_record[output_record_length++] = ';';
                 break;
        
      case '&':  output_record[output_record_length++] = '&';
                 output_record[output_record_length++] = 'a';
                 output_record[output_record_length++] = 'm';
                 output_record[output_record_length++] = 'p';
                 output_record[output_record_length++] = ';';
                 break;
        
      case '"':  output_record[output_record_length++] = '&';
                 output_record[output_record_length++] = 'q';
                 output_record[output_record_length++] = 'u';
                 output_record[output_record_length++] = 'o';
                 output_record[output_record_length++] = 't';
                 output_record[output_record_length++] = ';';
                 break;
        
      case '|':  output_record[output_record_length++] = '&';
                 output_record[output_record_length++] = 'b';
                 output_record[output_record_length++] = 'r';
                 output_record[output_record_length++] = 'v';
                 output_record[output_record_length++] = 'b';
                 output_record[output_record_length++] = 'a';
                 output_record[output_record_length++] = 'r';
                 output_record[output_record_length++] = ';';
                 break;
        
      case '\'': output_record[output_record_length++] = '&';
                 output_record[output_record_length++] = 'a';
                 output_record[output_record_length++] = 'p';
                 output_record[output_record_length++] = 'o';
                 output_record[output_record_length++] = 's';
                 output_record[output_record_length++] = ';';
                 break;

      /* normal characters */
      default:   output_record[output_record_length++] = f[i];
                 break;
    }
    /* mem checks */
    if (output_record_length > memseg_output_record)
      memadd_output_record(output_record_length);
  }

  /* close element tag */
  /* mem checks (18 = approx sizeof("</Lnnnnn") + sizeof("nnnnnnnn>\n") */
  if (output_record_length + t_length + 18 > memseg_output_record)
    memadd_output_record(output_record_length + t_length + 18);

  output_record[output_record_length++] = '<';
  output_record[output_record_length++] = '/';
  if (sw_xml_format == SW_SOAP)
  {
    strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
    output_record[output_record_length++] = ':';
  }
  if (sw_level_tag)
  {
    (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
    strcat_output_record((unsigned char *) temp_array);
  }
  strncat_output_record(t, t_length);
  if (n > 0)
  {
    if (sw_numbered_tag)
    {
      (void) sprintf(temp_array, "%ld", n);
      strcat_output_record((unsigned char *) temp_array);
    }
  }
  output_record[output_record_length++] = '>';
  output_record[output_record_length++] = '\n';
}


void mktag(tag_control, level, t, t_length, n, sw_numbered_tag, sw_numbered_attr, level_d, d, sw_delimiter_attr, sw_level_tag, sw_level_attr, sw_timestamp_attr, sw_custom_attr)
int tag_control;
int level;
unsigned char *t;
int t_length;
long n;
int sw_numbered_tag;
int sw_numbered_attr;
int level_d;
int d;
int sw_delimiter_attr;
int sw_level_tag;
int sw_level_attr;
int sw_timestamp_attr;
int sw_custom_attr;
{
  /* Construct an record tag array suitable for XML and UTF-8 printing */
  /* Array output_record and output_record_length are updated */
  long i;

  output_record_length = 0;

  /* padding */
  if (sw_xml_format == SW_SOAP)
  {
    for (i = 0 ; i < 4 ; i++)
      output_record[output_record_length++] = ' ';
  }
  for (i = 0 ; i < level ; i++)
  {
    output_record[output_record_length++] = ' ';
    output_record[output_record_length++] = ' ';
  }

  /* Not yet supported
   * if (tag_control == TAG_OPEN || tag_control == TAG_OPEN_CLOSE) */
  if (tag_control == TAG_OPEN)
  {
    /* open record tag */
    output_record[output_record_length++] = '<';
    if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      output_record[output_record_length++] = ':';
    }
    if (sw_level_tag)
    {
      (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
      strcat_output_record((unsigned char *) temp_array);
    }
    strncat_output_record(t, t_length);
    if (n > 0)
    {
      if (sw_numbered_tag)
      {
        (void) sprintf(temp_array, "%ld", n);
        strcat_output_record((unsigned char *) temp_array);
      }
      if (sw_numbered_attr)
      {
        (void) sprintf(temp_array, " %s=\"%ld\"", DEFAULT_STR_ATTRIBUTE_NUMBER, n);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
    if (sw_level_attr)
    {
      (void) sprintf(temp_array, " %s=\"%d\"", DEFAULT_STR_ATTRIBUTE_LEVEL, level);
      strcat_output_record((unsigned char *) temp_array);
    }
    if (sw_delimiter_attr)
    {
      if (d >= 0)
      {
        (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
        append_delimiter_string(level_d, d);
        output_record[output_record_length++] = '\"';
      }
      else if (d == DELIMITER_INDEX_WHITESPACE)
      {
        (void) sprintf(temp_array, " %s=\"WHITESPACE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
      }
      else if (d == DELIMITER_INDEX_NEWLINE)
      {
        (void) sprintf(temp_array, " %s=\"NEWLINE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
      }
      else
      {
        (void) sprintf(temp_array, " %s=\"\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
    if (sw_timestamp_attr)
    {
      time(&ts);
      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_DATE);
      strcat_output_record((unsigned char *) temp_array);
      make_timestamp_array(ts, TIMESTAMP_DATE);
      strcat_output_record((unsigned char *) temp_array);
      output_record[output_record_length++] = '\"';

      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_SECS);
      strcat_output_record((unsigned char *) temp_array);
      make_timestamp_array(ts, TIMESTAMP_SECS);
      strcat_output_record((unsigned char *) temp_array);
      output_record[output_record_length++] = '\"';
    }
    if (sw_custom_attr)
    {
      for (i = 0 ; i < nattrname ; i++)
      {
        if (attrname[i].level < 0 || attrname[i].level == level)
        {
          output_record[output_record_length++] = ' ';
          strncat_output_record(attrname[i].name, attrname[i].name_length);
          output_record[output_record_length++] = '=';
          output_record[output_record_length++] = '\"';
          strncat_output_record(attrname[i].value, attrname[i].value_length);
          output_record[output_record_length++] = '\"';
        }
      }
    }
  }
  else 
  {
    /* close record tag */
    output_record[output_record_length++] = '<';
    output_record[output_record_length++] = '/';
    if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      output_record[output_record_length++] = ':';
    }
    if (sw_level_tag)
    {
      (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
      strcat_output_record((unsigned char *) temp_array);
    }
    strncat_output_record(t, t_length);
    if (n > 0)
    {
      if (sw_numbered_tag)
      {
        (void) sprintf(temp_array, "%ld", n);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
  }
  /* 
   * Not yet supported
   * if (tag_control == TAG_OPEN_CLOSE)
   * {
   *   output_record[output_record_length++] = ' ';
   *   output_record[output_record_length++] = '/';
   * } 
   */
  output_record[output_record_length++] = '>';
  output_record[output_record_length++] = '\n';
}


void make_timestamp_array(t, type)
time_t t;
int type;
{
  /* put a timestamp into an array */
  int i;

  if (type == TIMESTAMP_SECS)
    (void) sprintf(temp_array, "%ld", t);
  else
  {
    (void) sprintf(temp_array, "%s", ctime(&t));
    /* strip new-line character from returned string */
    for (i = 0 ; i < strlen(temp_array) ; i++) 
      if (temp_array[i] == '\n') temp_array[i] = '\0';
  }
}

void append_delimiter_string(level, d)
int level;
int d;
{
  /* append delimiter string to output_record array and 
   * ensure printed delimiter doesn't corrupt XML output */
  int i;
  char hex_array[4];

  for (i = 0 ; i < delimiter[level]->length[d] ; i++)
  {
    /* only alphanum characters allowed */
    if (delimiter[level]->d[d][i] == '\n')
    {
      output_record[output_record_length++] = '\\';
      output_record[output_record_length++] = 'n';
    }
    else if (delimiter[level]->d[d][i] == '\r')
    {
      output_record[output_record_length++] = '\\';
      output_record[output_record_length++] = 'r';
    }
    else if (delimiter[level]->d[d][i] == '\t')
    {
      output_record[output_record_length++] = '\\';
      output_record[output_record_length++] = 't';
    }
    else if (delimiter[level]->d[d][i] == '\"' ||
             delimiter[level]->d[d][i] == '\'' ||
             delimiter[level]->d[d][i] == '`' ||
             delimiter[level]->d[d][i] == '<' ||
             delimiter[level]->d[d][i] == '>' ||
             delimiter[level]->d[d][i] == '&' ||
             delimiter[level]->d[d][i] == '|')
    {
      /* print HEX equivalent in URI style */
      (void) sprintf(hex_array, "%%%2.2x", delimiter[level]->d[d][i]);
      output_record[output_record_length++] = hex_array[0];
      output_record[output_record_length++] = toupper(hex_array[1]);
      output_record[output_record_length++] = toupper(hex_array[2]);
    }
    else if (isprint(delimiter[level]->d[d][i]))
      output_record[output_record_length++] = delimiter[level]->d[d][i];
    else
    {
      /* print HEX equivalent in URI style */
      (void) sprintf(hex_array, "%%%2.2x", delimiter[level]->d[d][i]);
      output_record[output_record_length++] = hex_array[0];
      output_record[output_record_length++] = toupper(hex_array[1]);
      output_record[output_record_length++] = toupper(hex_array[2]);
    }
  }
}


void strcpy_output_record(int level, const unsigned char *s_src)
{
  /* Write s_src to start of output_record array with padding */
  long i;

  output_record_length = 0;
  /* pad */
  while (output_record_length < level * 2)
    output_record[output_record_length++] = ' ';
  /* copy */
  for (i = 0 ; s_src[i] != '\0' ; i++)
    output_record[output_record_length++] = s_src[i];
}

void strncpy_output_record(const unsigned char *s_src, long n)
{
  /* Write at most n characters of s_src to start of output_record array */

  for (output_record_length = 0 ; output_record_length < n && s_src[output_record_length] != '\0' ; output_record_length++)
    output_record[output_record_length] = s_src[output_record_length];
}

void strcat_output_record(const unsigned char *s_src)
{
  /* Concatenate s_src to output_record array
   * starting from output_record_length. */
  long i;

  for (i = 0 ; s_src[i] != '\0' ; i++)
    output_record[output_record_length++] = s_src[i];
}

void strncat_output_record(const unsigned char *s_src, long n)
{
  /* Concatenate at most n characters of s_src to output_record array
   * starting from output_record_length. */
  long i;

  for (i = 0 ; i < n && s_src[i] != '\0' ; i++)
    output_record[output_record_length++] = s_src[i];
}

void format_key_value_tagname(i)
long i;
{
  /* neaten up what will become the new tag name */
  long j;

  /* re-trim field in case it was derived from a quoted field */
  trim_field(i);
  /* replace the whitespace between words with an underscore */
  for (j = input_record.field[i]->start ; j < input_record.field[i]->start + input_record.field[i]->length ; j++)
    if (isspace(input_record.line[j]))
      input_record.line[j] = '_';
}

