/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2009, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert command line text output into XML formatted output     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

void print_output_summary(argc, argv)
int argc;
unsigned char **argv;
{
  /* produce a summary tag */
  long i, j, k, record_cntr_processed;
  int ofd;

  /* set summary output stream */
  if (sw_summary_stderr)
    ofd = STDERR_FD;
  else
    ofd = STDOUT_FD;

  record_cntr_processed = 0;

  /* open summary tag */
  output_record_length = 0;
  output_record[output_record_length++] = '<';
  strcat_output_record(schema_parent_name);
  strcat_output_record((unsigned char *) "_summary>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* version */
  strcpy_output_record(1, (unsigned char *) "<version>");
  strcat_output_record((unsigned char *) VERSION);
  strcat_output_record((unsigned char *) "</version>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* timestamp */
  strcpy_output_record(1, (unsigned char *) "<timestamp>\n");
  utf8_print(ofd, output_record, output_record_length);
  make_timestamp_array(start_ts, TIMESTAMP_DATE);
  mkelem(2, (unsigned char *) "start", strlen("start"), (unsigned char *) temp_array, strlen(temp_array), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  utf8_print(ofd, output_record, output_record_length);
  time(&ts);
  make_timestamp_array(ts, TIMESTAMP_DATE);
  mkelem(2, (unsigned char *) "end", strlen("end"), (unsigned char *) temp_array, strlen(temp_array), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  utf8_print(ofd, output_record, output_record_length);
  strcpy_output_record(1, (unsigned char *) "</timestamp>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* open args tag */
  strcpy_output_record(1, (unsigned char *) "<args>\n");
  utf8_print(ofd, output_record, output_record_length);
  strcpy_output_record(2, (unsigned char *) "<command>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* command args */
  for (i = 0 ; i < argc ; i++)
  {
    mkelem(3, (unsigned char *) "arg", strlen("arg"), argv[i], xstrlen(argv[i]), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    utf8_print(ofd, output_record, output_record_length);
  }
  strcpy_output_record(2, (unsigned char *) "</command>\n");
  utf8_print(ofd, output_record, output_record_length);
  if (sw_schema != SCHEMA_DEFAULT)
  {
    strcpy_output_record(2, (unsigned char *) "<schema>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* schema args */
    for (i = 1 ; i < schema_argc ; i++)
    {
      mkelem(3, (unsigned char *) "arg", strlen("arg"), schema_argv[i], xstrlen(schema_argv[i]), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf8_print(ofd, output_record, output_record_length);
    }
    strcpy_output_record(2, (unsigned char *) "</schema>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  /* close args tag */
  strcpy_output_record(1, (unsigned char *) "</args>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* open separator tokens tag */
  strcpy_output_record(1, (unsigned char *) "<separator>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* delimiter tokens */
  for (i = 1 ; i < nlevels ; i++)
  {
    if (i == 1 && delimiter[i]->ndelimiters == 0)
    {
      (void) sprintf(temp_array, "<level%ld>\n", i);
      strcpy_output_record(2, (unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "      <delimiter>NEWLINE</delimiter>\n");
      (void) sprintf(temp_array, "    </level%ld>\n", i);
      strcat_output_record((unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
    }
    else if (i == 2 && delimiter[i]->ndelimiters == 0)
    {
      (void) sprintf(temp_array, "<level%ld>\n", i);
      strcpy_output_record(2, (unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "      <delimiter>WHITESPACE</delimiter>\n");
      (void) sprintf(temp_array, "    </level%ld>\n", i);
      strcat_output_record((unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
    }
    else if (delimiter[i]->ndelimiters > 0)
    {
      (void) sprintf(temp_array, "<level%ld>\n", i);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);

      for (j = 0 ; j < delimiter[i]->ndelimiters ; j++)
      {
        strcpy_output_record(3, (unsigned char *) "<delimiter");
        if (delimiter[i]->nscope[j])
        {
          strcat_output_record((unsigned char *) " scope=\"");
          for (k = 0 ; k < delimiter[i]->nscope[j] ; k++)
          {
            if (k)
              strcat_output_record((unsigned char *) ",");
            if (delimiter[i]->scope[j][k].min == delimiter[i]->scope[j][k].max)
              (void) sprintf(temp_array, "%ld", delimiter[i]->scope[j][k].min + 1);
            else if (delimiter[i]->scope[j][k].max < 0)
              (void) sprintf(temp_array, "%ld-MAX", delimiter[i]->scope[j][k].min + 1);
            else
              (void) sprintf(temp_array, "%ld-%ld", delimiter[i]->scope[j][k].min + 1, delimiter[i]->scope[j][k].max + 1);
            strcat_output_record((unsigned char *) temp_array);
          }
          strcat_output_record((unsigned char *) "\"");
        }
        strcat_output_record((unsigned char *) ">");
        append_delimiter_string(i, j);
        strcat_output_record((unsigned char *) "</delimiter>\n");
        utf8_print(ofd, output_record, output_record_length);
      }

      (void) sprintf(temp_array, "</level%ld>\n", i);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
    }
  }
  /* close separator tokens tag */
  strcpy_output_record(1, (unsigned char *) "</separator>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* column option */
  if (ncolumn)
  {
    /* open column tokens tag */
    strcpy_output_record(1, (unsigned char *) "<columns>\n");
    utf8_print(ofd, output_record, output_record_length);
    for (i = 0 ; i < ncolumn ; i++)
    {
      /* open field range tokens tag */
      (void) sprintf(temp_array, "<field%ld>\n", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
      /* open/close min tokens tag */
      if (column[i].min < 0)
        (void) sprintf(temp_array, "<min>LAST</min>\n");
      else
        (void) sprintf(temp_array, "<min>%ld</min>\n", column[i].min + 1);
      strcpy_output_record(3, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
      /* open/close max tokens tag */
      if (column[i].max < 0)
        (void) sprintf(temp_array, "<max>LAST</max>\n");
      else
        (void) sprintf(temp_array, "<max>%ld</max>\n", column[i].max + 1);
      strcpy_output_record(3, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
      /* open/close name tokens tag */
      if (column[i].name == NULL)
        strcpy_output_record(3, (unsigned char *) "<name>DEFAULT</name>\n");
      else
        mkelem(3, (unsigned char *) "name", strlen("name"), column[i].name, xstrlen(column[i].name), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf8_print(ofd, output_record, output_record_length);
      /* close field range tokens tag */
      (void) sprintf(temp_array, "</field%ld>\n", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
    }
    /* close column tokens tag */
    strcpy_output_record(1, (unsigned char *) "</columns>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  /* quotechars option */
  if (quotechars_length)
  {
    /* open quotechars tokens tag */
    strcpy_output_record(1, (unsigned char *) "<quotechars>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* quotechars tokens */
    for (i = 0 ; i < quotechars_length ; i++)
    {
      mkelem(2, (unsigned char *) "char", strlen("char"), &quotechars[i], 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf8_print(ofd, output_record, output_record_length);
    }
    /* close quotechars tokens tag */
    strcpy_output_record(1, (unsigned char *) "</quotechars>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  /* expel option */
  if (nexpel)
  {
    /* open expel tokens tag */
    strcpy_output_record(1, (unsigned char *) "<expel>\n");
    utf8_print(ofd, output_record, output_record_length);
    for (i = 0 ; i < nexpel ; i++)
    {
      /* open expel group tokens tag */
      (void) sprintf(temp_array, "<group%ld>\n", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
      /* open expel records tokens tag */
      strcpy_output_record(3, (unsigned char *) "<input_records>\n");
      utf8_print(ofd, output_record, output_record_length);
      for (j = 0 ; j < expel[i].n_erecords ; j++)
      {
        /* open range tokens tag */
        strcpy_output_record(4, (unsigned char *) "<range>\n");
        utf8_print(ofd, output_record, output_record_length);
        /* open/close type tokens tag */
        switch(expel[i].erecords[j].type)
        {
          case EXPEL_TYPE_N:
             strcpy_output_record(5, (unsigned char *) "<type>n</type>\n");
             break;
  
          case EXPEL_TYPE_F:
             strcpy_output_record(5, (unsigned char *) "<type>f</type>\n");
             break;

          case EXPEL_TYPE_C:
             strcpy_output_record(5, (unsigned char *) "<type>c</type>\n");
             break;
  
          default:
             (void) sprintf(temp_array, "<type>%d</type>\n", expel[i].erecords[j].type);
             strcpy_output_record(5, (unsigned char *) temp_array);
        }
        utf8_print(ofd, output_record, output_record_length);
        /* open/close min tokens tag */
        (void) sprintf(temp_array, "<min>%ld</min>\n", expel[i].erecords[j].min + 1);
        strcpy_output_record(5, (unsigned char *) temp_array);
        utf8_print(ofd, output_record, output_record_length);
        /* open/close max tokens tag */
        if (expel[i].erecords[j].max < 0)
          strcpy_output_record(5, (unsigned char *) "<max>LAST</max>\n");
        else
        {
          (void) sprintf(temp_array, "<max>%ld</max>\n", expel[i].erecords[j].max + 1);
          strcpy_output_record(5, (unsigned char *) temp_array);
        }
        utf8_print(ofd, output_record, output_record_length);
        /* open/close search string tokens tag */
        if (expel[i].erecords[j].s_length)
        {
          mkelem(5, (unsigned char *) "search", strlen("search"), expel[i].erecords[j].s, expel[i].erecords[j].s_length, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
          utf8_print(ofd, output_record, output_record_length);
        }
        /* close range tokens tag */
        strcpy_output_record(4, (unsigned char *) "</range>\n");
        utf8_print(ofd, output_record, output_record_length);
      }
      /* close expel records tokens tag */
      strcpy_output_record(3, (unsigned char *) "</input_records>\n");
      utf8_print(ofd, output_record, output_record_length);
      /* open expel fields tokens tag */
      strcpy_output_record(3, (unsigned char *) "<input_fields>\n");
      utf8_print(ofd, output_record, output_record_length);
      if (expel[i].n_efields)
      {
        for (j = 0 ; j < expel[i].n_efields ; j++)
        {
          /* open range tokens tag */
          strcpy_output_record(4, (unsigned char *) "<range>\n");
          utf8_print(ofd, output_record, output_record_length);
          /* open/close min tokens tag */
          if (expel[i].efields[j].min < 0)
            strcpy_output_record(5, (unsigned char *) "<min>LAST</min>\n");
          else
          {
            (void) sprintf(temp_array, "<min>%ld</min>\n", expel[i].efields[j].min + 1);
            strcpy_output_record(5, (unsigned char *) temp_array);
          }
          utf8_print(ofd, output_record, output_record_length);
          /* open/close max tokens tag */
          if (expel[i].efields[j].max < 0)
            strcpy_output_record(5, (unsigned char *) "<max>LAST</max>\n");
          else
          {
            (void) sprintf(temp_array, "<max>%ld</max>\n", expel[i].efields[j].max + 1);
            strcpy_output_record(5, (unsigned char *) temp_array);
          }
          utf8_print(ofd, output_record, output_record_length);
          /* close range tokens tag */
          strcpy_output_record(4, (unsigned char *) "</range>\n");
          utf8_print(ofd, output_record, output_record_length);
        }
      }
      else
      {
        /* open range tokens tag */
        strcpy_output_record(4, (unsigned char *) "<range>\n");
        utf8_print(ofd, output_record, output_record_length);
        /* open/close min tokens tag */
        strcpy_output_record(5, (unsigned char *) "<min>1</min>\n");
        utf8_print(ofd, output_record, output_record_length);
        /* open/close max tokens tag */
        strcpy_output_record(5, (unsigned char *) "<max>LAST</max>\n");
        utf8_print(ofd, output_record, output_record_length);
        /* close range tokens tag */
        strcpy_output_record(4, (unsigned char *) "</range>\n");
        utf8_print(ofd, output_record, output_record_length);
      }
      /* close expel fields tokens tag */
      strcpy_output_record(3, (unsigned char *) "</input_fields>\n");
      utf8_print(ofd, output_record, output_record_length);
      /* close expel group tokens tag */
      (void) sprintf(temp_array, "</group%ld>\n", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
    }
    /* close expel tokens tag */
    strcpy_output_record(1, (unsigned char *) "</expel>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  /* custom attribute option */
  if (nattrname)
  {
    /* open attrname tokens tag */
    strcpy_output_record(1, (unsigned char *) "<custom_attributes>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* attribute name tokens */
    for (i = 0 ; i < nattrname ; i++)
    {
      /* open attribute tokens tag */
      (void) sprintf(temp_array, "<attribute%ld>\n", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
      /* open/close level tokens tag */
      if (attrname[i].level < 0)
        strcpy_output_record(3, (unsigned char *) "<level>ALL</level>\n");
      else
      {
        (void) sprintf(temp_array, "<level>%d</level>\n", attrname[i].level);
        strcpy_output_record(3, (unsigned char *) temp_array);
      }
      utf8_print(ofd, output_record, output_record_length);
      /* open/close name tokens tag */
      mkelem(3, (unsigned char *) "name", strlen("name"), attrname[i].name, attrname[i].name_length, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf8_print(ofd, output_record, output_record_length);
      /* open/close value tokens tag */
      mkelem(3, (unsigned char *) "value", strlen("value"), attrname[i].value, attrname[i].value_length, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf8_print(ofd, output_record, output_record_length);
      /* close attribute tokens tag */
      (void) sprintf(temp_array, "</attribute%ld>\n", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf8_print(ofd, output_record, output_record_length);
    }
    /* close attrname tokens tag */
    strcpy_output_record(1, (unsigned char *) "</custom_attributes>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  /* Schema - open tag */
  strcpy_output_record(1, (unsigned char *) "<schema>\n");
  utf8_print(ofd, output_record, output_record_length);
  if (sw_schema == SCHEMA_DEFAULT)
  {
    /* Schema - default */
    strcpy_output_record(2, (unsigned char *) "<default>true</default>\n");
    utf8_print(ofd, output_record, output_record_length);
    record_cntr_processed = record_cntr_printed;
    /* Schema - open records tag */
    strcpy_output_record(2, (unsigned char *) "<records>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  else
  {
    /* Schema - not default */
    strcpy_output_record(2, (unsigned char *) "<default>false</default>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - file name */
    strcpy_output_record(2, (unsigned char *) "<file>");
    strcat_output_record(schema_filename);
    strcat_output_record((unsigned char *) "</file>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - type */
    if (sw_schema == SCHEMA_DTD)
      strcpy_output_record(2, (unsigned char *) "<interpreter>DTD</interpreter>\n");
    else if (sw_schema == SCHEMA_RNC)
      strcpy_output_record(2, (unsigned char *) "<interpreter>RNC</interpreter>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - open parent element */
    strcpy_output_record(2, (unsigned char *) "<parent>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - parent element name */
    strcpy_output_record(3, (unsigned char *) "<name>");
    strcat_output_record(schema_parent.name);
    strcat_output_record((unsigned char *) "</name>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - open elements tag */
    strcpy_output_record(3, (unsigned char *) "<elements>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - find all level 1 elements */
    for (i = 1 ; i < schema_parent.nelements ; i++)
    {
      for (k = 0 ; k < schema_parent.record->nfields ; k++)
      {
        if (xstrcmp(schema_element[i].name, schema_parent.record->field[k].name) == 0)
        {
          /* Schema - open element name tag */
          strcpy_output_record(4, (unsigned char *) "<");
          strcat_output_record(schema_element[i].name);
          strcat_output_record((unsigned char *) ">\n");
          utf8_print(ofd, output_record, output_record_length);
          /* Schema - match control */
          switch(schema_parent.record->field[k].match_control)
          { 
            case MATCH_ONE:          strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_ONE</match_control>\n"); break;
            case MATCH_NONE_OR_ONE:  strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_NONE_OR_ONE</match_control>\n"); break;
            case MATCH_NONE_OR_MANY: strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_NONE_OR_MANY</match_control>\n"); break;
            case MATCH_ONE_OR_MANY:  strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_ONE_OR_MANY</match_control>\n"); break;
          } 
          utf8_print(ofd, output_record, output_record_length);
          /* Schema - instances */
          for (j = 0 ; j < schema_element[i].nrecords ; j++)
          {
            /* Schema - open instance tag */
            (void) sprintf(temp_array, "<instance%ld>\n", j + 1);
            strcpy_output_record(5, (unsigned char *) temp_array);
            utf8_print(ofd, output_record, output_record_length);
            /* Schema - instance fields */
            (void) sprintf(temp_array, "<fields>%d</fields>\n", schema_element[i].record[j]->nfields);
            strcpy_output_record(6, (unsigned char *) temp_array);
            utf8_print(ofd, output_record, output_record_length);
            /* Schema - close instance tag */
            (void) sprintf(temp_array, "</instance%ld>\n", j + 1);
            strcpy_output_record(5, (unsigned char *) temp_array);
            utf8_print(ofd, output_record, output_record_length);
          }
          /* Schema - close element name tag */
          strcpy_output_record(4, (unsigned char *) "</");
          strcat_output_record(schema_element[i].name);
          strcat_output_record((unsigned char *) ">\n");
          utf8_print(ofd, output_record, output_record_length);
          /* element names are unique, don't compare any more */
          break;
        }
      }
    }
    /* Schema - close elements tag */
    strcpy_output_record(3, (unsigned char *) "</elements>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - close parent element */
    strcpy_output_record(2, (unsigned char *) "</parent>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - open records tag */
    strcpy_output_record(2, (unsigned char *) "<records>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - open elements tag */
    strcpy_output_record(3, (unsigned char *) "<elements>\n");
    utf8_print(ofd, output_record, output_record_length);
    /* Schema - elements */
    for (i = 1 ; i < schema_parent.nelements ; i++)
    {
      for (k = 0 ; k < schema_parent.record->nfields ; k++)
      {
        if (xstrcmp(schema_element[i].name, schema_parent.record->field[k].name) == 0)
        {
          /* Schema - open element name tag */
          strcpy_output_record(4, (unsigned char *) "<");
          strcat_output_record(schema_element[i].name);
          strcat_output_record((unsigned char *) ">\n");
          utf8_print(ofd, output_record, output_record_length);
          /* Schema - element total */
          (void) sprintf(temp_array, "<total>%d</total>\n", schema_element[i].element_count);
          strcpy_output_record(5, (unsigned char *) temp_array);
          utf8_print(ofd, output_record, output_record_length);
          /* Schema - element instance total */
          for (j = 0 ; j < schema_element[i].nrecords ; j++)
          {
            (void) sprintf(temp_array, "<instance%ld>%d</instance%ld>\n", j + 1, schema_element[i].record[j]->record_count, j + 1);
            strcpy_output_record(5, (unsigned char *) temp_array);
            utf8_print(ofd, output_record, output_record_length);
          }
          /* Schema - close element name tag */
          strcpy_output_record(4, (unsigned char *) "</");
          strcat_output_record(schema_element[i].name);
          strcat_output_record((unsigned char *) ">\n");
          utf8_print(ofd, output_record, output_record_length);
    
          record_cntr_processed += schema_element[i].element_count;

          break;
        }
      }
    }
    /* Schema - close elements tag */
    strcpy_output_record(3, (unsigned char *) "</elements>\n");
    utf8_print(ofd, output_record, output_record_length);
  }
  /* records read */
  (void) sprintf(temp_array, "<read>%ld</read>\n", record_cntr_read);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf8_print(ofd, output_record, output_record_length);
  /* records expelled */
  (void) sprintf(temp_array, "<expelled>%ld</expelled>\n", record_cntr_expelled);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf8_print(ofd, output_record, output_record_length);
  /* records matched */
  (void) sprintf(temp_array, "<matched>%ld</matched>\n", record_cntr_processed);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf8_print(ofd, output_record, output_record_length);
  /* records unmatched */
  (void) sprintf(temp_array, "<unmatched>%ld</unmatched>\n", record_cntr_unmatched);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf8_print(ofd, output_record, output_record_length);
  /* records blank */
  (void) sprintf(temp_array, "<ignoredblank>%ld</ignoredblank>\n", record_cntr_blank);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf8_print(ofd, output_record, output_record_length);
  /* records printed */
  (void) sprintf(temp_array, "<printed>%ld</printed>\n", record_cntr_printed);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf8_print(ofd, output_record, output_record_length);
  /* close records tag */
  strcpy_output_record(2, (unsigned char *) "</records>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* Schema - close tag */
  strcpy_output_record(1, (unsigned char *) "</schema>\n");
  utf8_print(ofd, output_record, output_record_length);
  /* close summary */
  strcpy_output_record(0, (unsigned char *) "</");
  strcat_output_record(schema_parent_name);
  strcat_output_record((unsigned char *) "_summary>\n");
  utf8_print(ofd, output_record, output_record_length);
}

